// Directmode Plasma
// Author: SkyBerron

//import femto.mode.HiRes16Color;
import femto.mode.Direct;
import femto.Game;
import femto.State;
import femto.input.Button;
import femto.font.TIC80;
//import static femto.hardware.ST7775.*;
import static femto.hardware.ST7775;
import lut512;

class Main extends State {
    static final var screen = new Direct(TIC80.font()); // the screenmode we want to draw with
    
    int counter; // variables are automatically initialized to 0 or null
    long timeLast = 0;
    
    int scene = 0;
    
    // Scanline
    int[] scanline;
    
    // Plasma palette
    int[] m_pal;
    
    // Look-up sin()
    static final int LUT_SIZE = 512;
    static final int LUT_MASK = 511;
    /*
    int[] m_sin[];
    */

    public static void main(String[] args){
        Game.run( TIC80.font(), new Main() );
    }
    
    void init(){
        screen.clear(0);
        
        // Scanline
        scanline = new int[220];
        
        // Plasma palette
        m_pal = new int[256];
        
        for( int i=0; i < 256; i++ ) {
            m_pal[i] = hsl( ( i * 360.0f ) / 256.0f, 1.0f, .5f );
        }
        /*
        m_sin = new int[LUT_SIZE];
        for( int i = 0; i < LUT_SIZE; i++ ) {
            m_sin[i] = (int) ( 255.5f * ( .5f + .5f* Math.sin( i * 2.0f * Math.PI / LUT_SIZE ) ) );
        }
        */
    }

    void update(){
        int halfw = screen.width()/2;
        int halfh = screen.height()/2;
        String txt = "noname";
        
        if( Button.A.justPressed() ){
            screen.clear(0);
            counter = 0;
        }
        if( Button.B.justPressed() ){
            screen.clear(0);
            scene = ( scene + 1 ) % 1;
        }
        
        if( scene == 0 ) {
            txt = "Plasma";

            /*
            int j = ( counter << 2 ) % 510;
            if( j > 255 ) j = 510 - j;
            float m = j / 256.0f;
            */
            
            pointer ptr = (pointer) lut512.bin();

            for( int y = 0; y < 176; y++ ) {
                /*
                for( int x = 0; x < 220; x++ ) {
                    scanline[x] = 0;
                }
                */
                /* SETTINGS 1
                float p1 = LUT_SIZE * ( 1.0f + Math.sin( ( y + counter ) / 100.0f ) );
                float dp1 = ( .5f * ( 1.0f + Math.sin( y / 30.0f ) ) * LUT_SIZE ) / 220.0f;
                float p2 = LUT_SIZE * ( 1.0f - Math.sin( y / 70.0f ) );
                float dp2 = ( .5f * ( 1.0f + Math.sin( y / 100.0f ) ) * LUT_SIZE ) / 220.0f;
                */
                /* SETTINGS 2
                float p1 = m_sin[ ( (int) ( 1.0f * y + ( counter << 2 ) ) ) & LUT_MASK ] * LUT_SIZE / 255.0f;
                float dp1 = m_sin[ ( (int) ( 1.2f * y ) ) & LUT_MASK ] * LUT_SIZE / ( 2.0f * 255.0f * 220.0f );
                float p2 = m_sin[ LUT_SIZE - 1 - ( (int) ( 1.1f * y ) ) & LUT_MASK ] * LUT_SIZE / 255.0f;
                float dp2 = m_sin[ ( (int) ( 1.3f * y ) ) & LUT_MASK ] * LUT_SIZE / ( 2.0f * 255.0f * 220.0f );
                */
                float p1 = ( (int) System.memory.LDRB( ptr + ( ( (int) ( 1.0f * y + ( counter << 2 ) ) ) & LUT_MASK ) ) ) * LUT_SIZE / 255.0f;
                float dp1 = ( (int) System.memory.LDRB( ptr + ( ( (int) ( 1.2f * y ) ) & LUT_MASK ) ) ) * LUT_SIZE / ( 2.0f * 255.0f * 220.0f );
                float p2 = ( (int) System.memory.LDRB( ptr + ( LUT_SIZE - 1 - ( ( (int) ( 1.1f * y ) ) & LUT_MASK ) ) ) ) * LUT_SIZE / 255.0f;
                float dp2 = ( (int) System.memory.LDRB( ptr + ( ( (int) ( 1.3f * y ) ) & LUT_MASK ) ) ) * LUT_SIZE / ( 2.0f * 255.0f * 220.0f );

                for( int x = 0; x < 220; x++ ) {
                    scanline[x] = m_pal[ ( ( (int) System.memory.LDRB( ptr + ( ( (int) p1 ) & LUT_MASK ) ) ) +
                        ( (int) System.memory.LDRB( ptr + ( ( (int) p2 ) & LUT_MASK ) ) ) ) & 255 ];
                    p1 += dp1;
                    p2 += dp2;
                }
                drawScanline( 0, y, 220, scanline, 0, 220 );
            }
        }

        if( true ) {
            // Print some text
            int y = screen.height() - 6;
            if( scene > 0 ) {
                screen.fillRect( 0, y, screen.width(), 6, 0 );
            }
            screen.setTextPosition( 0, y );
            screen.textColor++;
            screen.println( "Scene " + scene + ": " + txt );
            screen.setTextPosition( screen.width() - 32, y );
            screen.println( (System.currentTimeMillis() - timeLast) + " ms " );
        }

        counter++;
        timeLast = System.currentTimeMillis();
 
    }
    
    // no built-in abs() for floats
    public float abs( float a )   {
        return( ( a >= 0.0f ) ? a : (-a) );
    }
    
    // no built-in % operator for floats
    public float mod( float a, float b ) {
        // using int, casting to long not working
        int d = (int) ( a / b );
        return( a - b * d );
    }
    
    public float mid( float a, float b, float c ) {
        float d = a;
        if( a < b ) {
            if( b < c ) {
                d = b;
            } else if( a < c ) {
                d = c;
            }
        } else if( c < a ) {
            if( b < c ) {
                d = c;
            } else {
                d = b;
            }
        }
        return d;
    }
    
    public int mid( int a, int b, int c ) {
        int d;
        if( a > b ) {
            d = a; a = b; b = d;
        }
        if( b > c ) {
            d = b; b = c; c = d;
        }
        if( a > b ) {
            return a;
        }
        return b;
    }
    
    // no built-in hsl color space conversion
    public int hsl( float h, float s, float l )
    {
        h = mod( h, 360.0f );
        float c = ( 1.0f - abs( 2.0f*l - 1.0f ) ) * s;
        float x = c * ( 1.0f - abs( mod ( h / 60.0f, 2.0f ) - 1.0f ) );
        float m = l - c / 2.0f;
        
        int u = (int) ( .5f + ( c + m ) * 255.0f );
        int v = (int) ( .5f + ( x + m ) * 255.0f );
        
        int r = 0, g = 0, b = 0;
        
        if( h < 180.0f ) {
            if( h < 60.0f ) {
                r = u; g = v;
            } else if( h < 120.0f ) {
                r = v; g = u;
            } else {
                g = u; b = v;
            }
        } else {
            if( h < 240.0f ) {
                g = v; b = u;
            } else if( h < 300.0f ) {
                r = v; b = u;
            } else {
                r = u; b = v;
            }
        }
        
        return (((int)(r>>3))<<11)
                | (((int)(g>>2))<<5)
                | ((int)(b>>3));
    }

    public static void drawScanline( int x, int y, int w, int[] s, int xs, int xf ) {
        if( xs < 0 ) {
            x -= xs;
            w += xs;
            xs = 0;
        }

        if( x < 0 ) {
            xs -= x;
            w += x;
            x = 0;
        }
        
        if( xs + w >= xf ) {
            w = xf - xs;
        }
        
        if( x + w >= 220 )
            w = 220 - x;
        
        if( y < 0 || y >= 176 || x >= 220 || w <= 0 )
            return;

        ST7775.setX(x);
        ST7775.setY(y);
        ST7775.beginStream();
        while( w-- > 0 ) {
            ST7775.writeData( s[xs++] );
        }
        ST7775.setX(0);
        ST7775.setY(0);
        
    }
    
    public void dorun( int[] s, int w, int x0, int x1, int v ) {
        if( x0 < 0 )
            x0 = 0;
        if( x1 >= w )
            x1 = w - 1;
        for( int i = x0; i <= x1; i++ ) {
            s[i] = v;
        }
    }

}
